import sys
import time
import boto3
from io import StringIO
from datetime import datetime
from botocore.exceptions import ClientError
from awsglue.utils import getResolvedOptions
from traffic_metrics.utils import get_secret
from traffic_metrics.utils import check_if_ts_table_exists
from sqlenv.wavefront.wavefront_sender_factory import WavefrontSenderFactory
from traffic_metrics.traffic_metrics_cloudwatch_export import (
    print_traffic_stats_cloudwatch,
)
from traffic_metrics.traffic_metrics_timestream_export import (
    print_traffic_stats_timestream,
)

default_tags = {
    "service": "traffic-metrics",
    "application": "traffic-metrics",
    "runtime": "SPARK",
    "shard": "none",
    "sguid": "psc:cdca:data_platform:traffic-metrics",
}

environments_to_export = ["prd1ew2", "int1usw2", "scl2use2"]

if __name__ == "__main__":
    if "--{}".format("end_time") in sys.argv:
        args = getResolvedOptions(
            sys.argv,
            ["JOB_NAME", "environment", "region", "end_time", "environment_category"],
        )
        env = args["environment"]
        region = args["region"]
        end_time = args["end_time"]
        env_type = args["environment_category"]
        dt = datetime.fromtimestamp(end_time)
    else:
        args = getResolvedOptions(
            sys.argv, ["JOB_NAME", "environment", "region", "environment_category"]
        )
        env = args["environment"]
        region = args["region"]
        env_type = args["environment_category"]
        end_time = time.time()
        dt = datetime.fromtimestamp(end_time)

    default_tags.update({"awsRegion": region})
    default_tags.update({"env": env_type})

    wf_sender_factory = WavefrontSenderFactory(
        api_token=get_secret("mcs2/" + env + "/wavefront-token", "us-west-2"),
        default_tags=default_tags,
        project_name="traffic-metrics",
        url_suffix="sbu.wavefront.com",
    )

    wf_sender_standard = wf_sender_factory.create_and_get_wavefront_sender("STANDARD")

    if env in environments_to_export:
        buffer = StringIO()
        sys.stdout = buffer

        if (
            check_if_ts_table_exists(env + "-psc-lc-metrics", env + "-org-metrics")
            == True
        ):
            print_traffic_stats_timestream(end_time, 1, env, region)
        else:
            print_traffic_stats_cloudwatch(end_time, 1, env, region)

        export_output = buffer.getvalue()
        sys.stdout = sys.__stdout__

        print(export_output)

        date = {"year": dt.year, "month": dt.month, "day": dt.day}

        s3 = boto3.resource("s3")

        month = "0" + str(date["month"]) if date["month"] < 10 else str(date["month"])
        day = "0" + str(date["day"]) if date["day"] < 10 else str(date["day"])

        key_name = (
            "traffic_snapshots/year="
            + str(date["year"])
            + "/month="
            + month
            + "/day="
            + day
            + "/"
            + region
            + ".tsv"
        )

        if env == "prd1ew2":
            export_bucket_name = "cdca-prod-data-at-scale-incoming"
        else:
            export_bucket_name = "cdca-prod-data-at-scale-incoming-dev"

        object = s3.Object(bucket_name=export_bucket_name, key=key_name)

        try:
            object.put(Body=export_output, ACL="bucket-owner-full-control")
            wf_sender_standard.send_job_success_metric(
                env + "-traffic-metrics-infra-glue-job"
            )

        except ClientError as error:
            wf_sender_standard.send_job_failure_metric(
                env + "-traffic-metrics-infra-glue-job"
            )
            raise error

        finally:
            wf_sender_standard.close_wf_client_connection()
