import re

import boto3
import logging
import time
import sys
import dateutil.parser
from datetime import datetime
from psc_orgkey import decode_orgkey
from itertools import zip_longest
from botocore.config import Config

_logger = logging.getLogger(__name__)
MIN_ACTIVE_ENDPOINTS = 1
MIN_TRAFFIC = 1
METRIC_PERIOD_S = 3600


class MetricReader:
    def __init__(
        self, env_id, look_back_metric_points, metric_period, end_time, region_name
    ):
        config = Config(retries=dict(max_attempts=20))
        self.cw = boto3.client("cloudwatch", region_name=region_name, config=config)
        self.ts = boto3.client("timestream-query", region_name=region_name)

        self.namespace = env_id + "_custom"
        self.look_back_metric_points = look_back_metric_points
        self.metric_period = metric_period
        self.start_time = datetime.utcfromtimestamp(
            end_time - (metric_period * look_back_metric_points)
        )
        self.end_time = datetime.utcfromtimestamp(end_time)
        self.cache_timeseries = {}
        self.database = "psc-lc-metrics"
        self.table = "org-metrics"
        self.env = env_id

    def _get_dimensions(self, metric_name):
        _logger.info(f"Getting Dimensions for {metric_name}")
        dimensions = []
        paginator = self.cw.get_paginator("list_metrics")
        for response in paginator.paginate(
            Namespace=self.namespace, MetricName=metric_name
        ):
            _logger.debug(f"Getting next set of dimensions")
            for metric in response["Metrics"]:
                dimensions += metric["Dimensions"]
        return dimensions

    def get_metric_series(self, metric_name, aws_stat):
        _logger.info(f"Getting Metric series for {metric_name}")
        metric_id = metric_name + "-" + aws_stat
        if metric_id in self.cache_timeseries:
            return self.cache_timeseries[metric_id]

        dimensions = self._get_dimensions(metric_name)
        if len(dimensions) == 0:
            return {}
        while True:
            try:
                series = self._metric_get_helper(dimensions, metric_name, aws_stat)
                self.cache_timeseries[metric_id] = series
                return series
            except Exception as ex:
                _logger.warning(f"Exception getting metrics from CW: {ex}")
                time.sleep(5)

    def _metric_get_helper(self, dimensions, metric_name, aws_stat):
        series = {}
        group_dimensions = self._grouper(dimensions, 100)
        for hundred_dimensions in group_dimensions:
            request_payload = []
            for dimension in hundred_dimensions:
                if dimension:
                    #  Defensive code, in dev we had some bad org names with dashes,  That shouldn't be a valid orgkey
                    #  And is invalid to insert into the Id field here
                    if "-" not in dimension["Value"]:
                        metric_request = {
                            "Id": f"a_{dimension['Value']}",
                            "MetricStat": {
                                "Metric": {
                                    "Namespace": self.namespace,
                                    "MetricName": metric_name,
                                    "Dimensions": [dimension],
                                },
                                "Period": self.metric_period,
                                "Stat": aws_stat,
                            },
                        }
                        request_payload.append(metric_request)
            paginator = self.cw.get_paginator("get_metric_data")
            for response in paginator.paginate(
                MetricDataQueries=request_payload,
                StartTime=self.start_time,
                EndTime=self.end_time,
            ):
                _logger.debug(f"Getting next set of DataPoints")
                self._parse_metric_results(series, response)
        return series

    @staticmethod
    def _parse_metric_results(series, response):
        for results in response["MetricDataResults"]:
            if results["StatusCode"] == "Complete":
                #  Split off the `a_` we prepended to org ID in the Id field above.  AWS requires first character to be a letter
                org = results["Id"].split("_")[1]
                series[org] = {
                    "Timestamps": results["Timestamps"],
                    "Values": results["Values"],
                }

    @staticmethod
    def _grouper(inputs, n, fill_value=None):
        # return first n as a tuple, then next n, etc.  Pad final set with None
        iters = [iter(inputs)] * n
        return zip_longest(*iters, fillvalue=fill_value)

    def get_sum_series(self, metric_name, aws_stat, extrapolate=False):
        series = self.get_metric_series(metric_name, aws_stat)
        sum_series = {}
        for org in series:
            last_x_values = series[org]["Values"][-self.look_back_metric_points :]
            sum_org = sum(last_x_values)
            number_points = len(last_x_values)
            if extrapolate and 0 < number_points < self.look_back_metric_points:
                sum_org *= 1.0 * self.look_back_metric_points / number_points
            sum_series[org] = sum_org
        return sum_series

    def get_sum_series_highest_x_points(
        self, metric_name, aws_stat, pro_rate, top_x_points
    ):
        # Return sum of the top x metric points over the look_back_metric_points range.
        # Useful if you want to scale to highest data points over a long period of time.
        # So that you don't churn up and down a lot.
        series = self.get_metric_series(metric_name, aws_stat)
        highest_x_series = {}
        for org in series:
            top_x_values = sorted(series[org]["Values"])[-top_x_points:]
            highest_x_org = sum(top_x_values)
            number_points = len(top_x_values)
            if pro_rate:
                # If we don't get top_x_points pro rate out to top x points.
                if 0 < number_points < top_x_points:
                    highest_x_org *= 1.0 * top_x_points / number_points
            highest_x_series[org] = highest_x_org
            _logger.debug(f"org: {org}, sum: {highest_x_org}, values; {top_x_values}")
        return highest_x_series

    def get_avg_series(self, metric_name, aws_stat):
        series = self.get_metric_series(metric_name, aws_stat)
        avg_series = {}
        sum_all = 0
        num_all = 0
        for org in series:
            number_values = len(series[org]["Values"])
            if number_values == 0:
                avg_series[org] = 0
            else:
                sum_org = sum(series[org]["Values"])
                sum_all += sum_org / number_values
                num_all += 1
                avg_series[org] = sum_org / number_values
        return avg_series

    def get_sum_series_ts(
        self, identifier_column, metric_name, aws_stat, extrapolate=False
    ):
        series = self.get_metric_series_ts(identifier_column, metric_name, aws_stat)
        sum_series = {}
        for org in series:
            last_x_values = series[org][-self.look_back_metric_points :]
            sum_org = sum(last_x_values)
            number_points = len(last_x_values)
            if extrapolate and 0 < number_points < self.look_back_metric_points:
                sum_org *= 1.0 * self.look_back_metric_points / number_points
            sum_series[org] = sum_org
        return sum_series

    def get_avg_series_ts(self, identifier_column, metric_name, aws_stat):
        series = self.get_metric_series_ts(identifier_column, metric_name, aws_stat)
        avg_series = {}
        sum_all = 0
        num_all = 0
        for org in series:
            number_values = len(series[org])
            if number_values == 0:
                avg_series[org] = 0
            else:
                sum_org = sum(series[org])
                sum_all += sum_org / number_values
                num_all += 1
                avg_series[org] = sum_org / number_values
        return avg_series

    def get_metric_series_ts(self, identifier_column, metric_name, aws_stat):
        query = self.build__query(identifier_column, metric_name, aws_stat)
        paginator = self.ts.get_paginator("query")
        metric_series = {}
        retries = 0
        while True:
            try:
                page_iterator = paginator.paginate(QueryString=query)
                for page in page_iterator:
                    MetricReader.get_query_status(page, query)
                    self.parse_query_result(page, metric_series)
                    _logger.debug(f"Success executing last_x query for {metric_name}")
            except Exception as ex:
                MetricReader.log_and_sleep(
                    f"Timestream Query Exception: {ex} retrieving data for {metric_name}, Retrying",
                    retries,
                    max_sleep=60,
                )
            else:
                break
            retries += 1

        return metric_series

    @staticmethod
    def parse_query_result(query_result, metric_series):
        for row in query_result["Rows"]:
            data = row["Data"]
            identifier = data[0]["ScalarValue"]

            if bool(re.match("^[^_-]+$", identifier)):
                if metric_series.get(identifier) is None:
                    metric_series[identifier] = []

                if not data[2].get("NullValue"):
                    metric_series[identifier].append(float(data[2]["ScalarValue"]))

        return metric_series

    def build__query(self, identifier_column, metric_name, aws_stat):
        return f"""
        SELECT {identifier_column}, bin(time, {str(self.metric_period)}s) as hour,
        {aws_stat}({metric_name}) as datapoints
        FROM \"{self.env}-{self.database}\".\"{self.env}-{self.table}\"
        WHERE time BETWEEN '{str(self.start_time)}' and '{str(self.end_time)}'
        GROUP BY {identifier_column}, bin(time, {str(self.metric_period)}s)  ORDER BY 1 ASC, 2 DESC
        """

    @staticmethod
    def get_query_status(query_result, query):
        query_status = query_result["QueryStatus"]
        progress_percentage = query_status["ProgressPercentage"]
        _logger.debug(
            "Query progress for Query:\n"
            + query
            + "\n so far: "
            + str(progress_percentage)
            + "%"
        )
        return progress_percentage

    def log_and_sleep(message, retries=0, base_sleep=0.1, max_sleep=120):
        """
        Utility function to use for exponential backoff when retrying
        """
        sleep_time = min((2**retries) * base_sleep, max_sleep)
        if sleep_time < max_sleep:
            _logger.debug(message)
        else:
            _logger.error(message)

        time.sleep(sleep_time)
        return message


def print_traffic_stats_timestream(end_time, days_back, env_id, region_name):
    traffic_multiplier = 1.0 / days_back / 1024 / 1024
    hours_back = days_back * 24

    data_rate_metric_reader = MetricReader(
        env_id, hours_back, METRIC_PERIOD_S, end_time, region_name
    )

    cbth_traffic_ts = data_rate_metric_reader.get_sum_series_ts(
        "org_key", "ingress_data_cbth", "Sum"
    )
    cbd_traffic_ts = data_rate_metric_reader.get_sum_series_ts(
        "org_key", "ingress_data_cbd", "Sum"
    )
    cbth_events_ts = data_rate_metric_reader.get_sum_series_ts(
        "org_key", "ingress_events_cbth", "Sum"
    )
    cbd_events_ts = data_rate_metric_reader.get_sum_series_ts(
        "org_key", "ingress_events_cbd", "Sum"
    )
    cbd_events_longterm_ts = data_rate_metric_reader.get_sum_series_ts(
        "org_key", "ingress_events_cbd_long_term", "Sum"
    )

    windows_traffic_ts = data_rate_metric_reader.get_sum_series_ts(
        "org_key", "ingress_data_windows", "Sum"
    )
    linux_traffic_ts = data_rate_metric_reader.get_sum_series_ts(
        "org_key", "ingress_data_linux", "Sum"
    )
    mac_traffic_ts = data_rate_metric_reader.get_sum_series_ts(
        "org_key", "ingress_data_mac", "Sum"
    )
    indexed_data_cw = data_rate_metric_reader.get_sum_series("indexed_data", "Sum")

    # Go 3 days back since some metrics are sporradic (especially for small envs)
    endpoint_count_metric_reader = MetricReader(
        env_id, 72, METRIC_PERIOD_S, end_time, region_name
    )

    active_endpoints_ts = endpoint_count_metric_reader.get_avg_series_ts(
        "org_key", "active_endpoints", "max"
    )
    active_endpoints_cbd_ts = endpoint_count_metric_reader.get_avg_series_ts(
        "org_key", "active_endpoints_cbd", "max"
    )
    active_endpoints_linux_ts = endpoint_count_metric_reader.get_avg_series_ts(
        "org_key", "active_endpoints_linux", "max"
    )
    active_endpoints_windows_ts = endpoint_count_metric_reader.get_avg_series_ts(
        "org_key", "active_endpoints_windows", "max"
    )
    active_endpoints_mac_ts = endpoint_count_metric_reader.get_avg_series_ts(
        "org_key", "active_endpoints_mac", "max"
    )
    active_endpoints_cbd_linux_ts = endpoint_count_metric_reader.get_avg_series_ts(
        "org_key", "active_endpoints_cbd_linux", "max"
    )
    active_endpoints_cbd_windows_ts = endpoint_count_metric_reader.get_avg_series_ts(
        "org_key", "active_endpoints_cbd_windows", "max"
    )
    active_endpoints_cbd_mac_ts = endpoint_count_metric_reader.get_avg_series_ts(
        "org_key", "active_endpoints_cbd_mac", "max"
    )

    print(
        f"org_key\tendpoints\tendpoints_windows\tendpoints_linux\tendpoints_mac\tendpoints_cbd\ttraffic_windows\ttraffic_linux\ttraffic_mac\ttraffic_cbth\ttraffic_cbd\tindexed_data\tevents_cbth\tevents_cbd\tevents_longterm_cbd\tenv\torg_id\tactive_endpoints_cbd_windows\tactive_endpoints_cbd_linux\tactive_endpoints_cbd_mac"
    )
    for k, v in sorted(active_endpoints_ts.items()):
        traffic_cbth_ts = cbth_traffic_ts.get(k, 0) * traffic_multiplier
        traffic_cbd_ts = cbd_traffic_ts.get(k, 0) * traffic_multiplier
        if v < MIN_ACTIVE_ENDPOINTS and traffic_cbth_ts + traffic_cbd_ts < MIN_TRAFFIC:
            continue

        decoded = decode_orgkey(orgkey=k)
        print(
            f"{k}\t{int(v)}\t{int(active_endpoints_windows_ts.get(k, 0))}\t{int(active_endpoints_linux_ts.get(k, 0))}\t"
            f"{int(active_endpoints_mac_ts.get(k, 0))}\t{int(active_endpoints_cbd_ts.get(k, 0))}\t"
            f"{windows_traffic_ts.get(k, 0) * traffic_multiplier:0.2f}\t{linux_traffic_ts.get(k, 0) * traffic_multiplier:0.2f}\t{mac_traffic_ts.get(k, 0) * traffic_multiplier:0.2f}\t"
            f"{traffic_cbth_ts:0.2f}\t{traffic_cbd_ts:0.2f}\t{indexed_data_cw.get(k, 0) * traffic_multiplier:0.2f}\t"
            f"{int(cbth_events_ts.get(k, 0))}\t{int(cbd_events_ts.get(k, 0))}\t{int(cbd_events_longterm_ts.get(k, 0))}\t{decoded.env}\t{decoded.org_id}\t"
            f"{int(active_endpoints_cbd_windows_ts.get(k, 0))}\t{int(active_endpoints_cbd_linux_ts.get(k, 0))}\t{int(active_endpoints_cbd_mac_ts.get(k, 0))}"
        )
